# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testbed module."""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

logger = gdm_logger.get_logger()

COMMANDS = {"FILE_SYSTEMS": "df"}


class Testbed(object):
  """Class for testbed interaction."""

  def __init__(self, devices, testing_props, name="testbed"):
    """Initialize a Testbed object.

    Args:
        devices (list): list of device instances.
        testing_props (dict): properties of the testbed used for testing.
        name (str): optional name of the testbed, defaulted to "testbed".
    """
    self.name = name
    self.devices = devices
    self.testing_props = testing_props
    self.commands = COMMANDS

    # map from testing prop to corresponding testbed health check method
    # e.g. if { "external_storage": self.check_external_storage_mounted }, then the
    #      check_external_storage_mounted method will be executed as a health check
    #      if the "external_storage" prop is provided during Testbed initialization.
    self._PROP_TO_HEALTH_CHECK = {
        "external_storage": self.check_external_storage_mounted
    }

  @decorators.health_check
  def check_external_storage_mounted(self, enabled):
    """Check if external USB storage is mounted on devices in the testbed.

    Args:
        enabled (str): whether to run this health check, "true" or "false".

    Raises:
        TestbedMissingExternalStorageError: if devices are missing external
        USB storage.
    """
    if enabled.lower() == "false":
      return

    file_system_cmd = self.commands["FILE_SYSTEMS"]
    usb_storage_location = "/data/usb"
    devices_missing_external_storage = []
    for device in self.devices:
      file_system_output = device.shell(file_system_cmd)
      file_system_mounts = self._get_filesystem_mounts(file_system_output)
      if usb_storage_location not in file_system_mounts:
        devices_missing_external_storage.append(device.name)

    if devices_missing_external_storage:
      raise errors.TestbedMissingExternalStorageError(
          device_names=devices_missing_external_storage,
          details=(
              f"command {file_system_cmd!r} did not display any filesystems "
              f"mounted on {usb_storage_location!r}"),
          recovery="mount external USB storage")

  @decorators.LogDecorator(logger)
  def check_testbed_ready(self):
    """Checks if the testbed is healthy by executing a series of health check methods."""
    health_check_call = []  # list of tuples: (health check method, prop value)
    for prop in self.testing_props:
      if prop in self._PROP_TO_HEALTH_CHECK:
        health_check_call.append(
            (self._PROP_TO_HEALTH_CHECK[prop], self.testing_props[prop]))
    self._execute_health_check_methods(health_check_call)

  @decorators.LogDecorator(logger)
  def make_testbed_ready(self, setting="on"):
    """Check testbed readiness and attempt recovery if allowed.

    Args:
        setting (str): 'on'|'off'|'check_only'.

    Raises:
        DeviceError: if testbed health checks fail.
    """
    if setting not in ["on", "off", "check_only"]:
      raise errors.DeviceError(
          "{} failed make_testbed_ready. Setting '{}' is invalid.".format(
              self.name, setting))
    elif setting == "off":
      return

    # check testbed health
    try:
      self.check_testbed_ready()
    except errors.DeviceError as err:
      if setting == "check_only":
        logger.info("{} skipping recovery".format(self.name))
        raise

      logger.info("{} failed check_testbed_ready with {}".format(
          self.name, repr(err)))

      # attempt to recover the testbed
      self.recover(err)
      logger.info("{} re-checking readiness after recovery attempt".format(
          self.name))
      self.check_testbed_ready()
      logger.info("{} successfully recovered to ready state".format(self.name))

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Attempts to recover testbed based on the type of error specified.

    Args:
        error (CheckTestbedReadyError): The exception thrown that will be
          used to identify a possible recovery solution to use.

    Raises:
        CheckTestbedReadyError: If recovery fails.
    """
    raise error

  def _execute_health_check_methods(self, health_check_calls):
    """Execute health checks on the testbed.

    Args:
        health_check_calls (list): list of tuples as (health check method,
          argument) to execute as health checks.

    Raises:
        CheckTestbedReadyError: if health check fails.

    Notes:
        - The value of the property that triggered the health check is
        passed to the health
        check method as an argument; therefore, health check methods will
        accept a single
        argument (unlike device health checks). All property values will be
        strings.
        - Some health checks will have "true" / "false" argument values. The
        health check
        should be skipped when the argument is "false".
        - Order of health check methods matter. Health check execution will
        stop at the first
        failing health check.
    """
    checks_passed = []
    for step, health_check_call in enumerate(health_check_calls):
      health_check_method, arg = health_check_call
      method_name = health_check_method.__name__
      health_check_name = method_name.replace("_", " ").strip().capitalize()

      try:
        health_check_method(arg)
      except errors.CheckTestbedReadyError as err:
        logger.info("testbed health check {}/{} failed: {}.".format(
            step + 1, len(health_check_calls), health_check_name))
        err.checks_passed = checks_passed
        err.properties = self.testing_props.copy()
        raise

      checks_passed.append("{}.{}".format(type(self).__name__, method_name))
      logger.info("testbed health check {}/{} succeeded: {}.".format(
          step + 1, len(health_check_calls), health_check_name))

  def _get_filesystem_mounts(self, file_system_output):
    """Parse locations that file systems are mounted on from command output.

    Args:
        file_system_output (str): output of command listing device file
          systems.

    Returns:
        list: locations that file systems are mounted on on the device.
    """
    file_system_mounts = []
    for file_system in file_system_output.splitlines():
      if not file_system:
        continue
      file_system_info = file_system.split()

      # 6 columns of information
      # Filesystem,1K-blocks,Used,Available,Use%,Mounted on
      if len(file_system_info) == 6:
        file_system_mounts.append(file_system_info[5])

    return file_system_mounts



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device class for EFR32 Pigweed Lighting device."""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device.base_classes import silabs_efr32_device
from gazoo_device.capabilities import pwrpc_light_default
from gazoo_device.utility import pwrpc_utils

# TODO(b/185956488): Remove conditional imports of Pigweed
try:
  # pylint: disable=g-import-not-at-top
  from button_service import button_service_pb2
  from device_service import device_service_pb2
  from lighting_service import lighting_service_pb2
except ImportError:
  button_service_pb2 = None
  device_service_pb2 = None
  lighting_service_pb2 = None


class EFR32PigweedLighting(silabs_efr32_device.SilabsEFR32Device):
  """Device class for EFR32PigweedLighting devices.

  This device class is for the Pigweed lighting application running on
  the Silabs EFR32 platform:
  https://github.com/project-chip/connectedhomeip/blob/master/examples/lighting-app/efr32/README.md
  """
  DETECT_MATCH_CRITERIA = {
      detect_criteria.PigweedQuery.product_name: "j-link",
      detect_criteria.PigweedQuery.manufacturer_name: "silicon_labs",
      detect_criteria.PigweedQuery.app_type:
          pwrpc_utils.PigweedAppType.LIGHTING.value,
  }
  DEVICE_TYPE = "efr32pigweedlighting"
  _OWNER_EMAIL = "gdm-authors@google.com"
  _COMMUNICATION_KWARGS = {"protobufs": (button_service_pb2,
                                         lighting_service_pb2,
                                         device_service_pb2),
                           "baudrate": silabs_efr32_device.BAUDRATE}

  @decorators.DynamicProperty
  def firmware_version(self):
    """Firmware version of the device."""
    # TODO(b/185298972)
    # Add firmware version endpoint once the pull request is merged:
    # https://github.com/project-chip/connectedhomeip/pull/5866
    return "0"

  @decorators.CapabilityDecorator(pwrpc_light_default.PwRPCLightDefault)
  def pw_rpc_light(self):
    """PwRPCLight instance to send RPC commands."""
    return self.lazy_init(pwrpc_light_default.PwRPCLightDefault,
                          device_name=self.name,
                          switchboard_call=self.switchboard.call)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.

Returns a dictionary mapping a persistent address (adb or serial port) to usb
info for that address.

Note: usb_info_linux primarily relies on pyudev for full information.
It processes just adb devices and those that show up in /dev/serial/by-id
"""
import re
import sys

from gazoo_device.utility import usb_config
import pyudev


PERSISTENT_SERIAL_PATH_FOLDER = '/dev/serial/by-id'
MOUNTED_DISK_FOLDER = '/dev/disk/by-id'
ADB_SERIAL_REGEX = r'\/dev\/disk\/by-id\/usb-Linux_File-CD_Gadget_([\da-z]+)-0:0'
IDENTIFIER_REGEX = r'\d-\d(?:\.\d?)+(\d\.\d):\d\.\d'
ANDROID_DEVICE_PRODUCT_NAME = 'File-CD Gadget'

# These vendor's mounted drives show up as multiple usb connections.
# We use the one with the symlink.
REPEAT_VENDORS = [
    'FTDI',
    'cambrionix',
    'Linux',
    'Android',
]


def get_address_to_usb_info_dict():
  """Returns a dict of address to the usb information.

  Returns:
     dict: dictionary of address to UsbInfo instance for each usb device
     attached.
  """
  udev_devices = get_pyudev_list_of_devices()
  relevant_devices = [
      device for device in udev_devices if device.get('ID_BUS') == 'usb'
  ]
  address_to_usb_info_dict = {}
  location_dict = {}
  model_dict = {}
  for device in relevant_devices:
    address, entry, dev_path, model = _process_udev_device(device)
    if address:  # remove devices that don't have good addresses
      address_to_usb_info_dict[address] = entry
      if dev_path:
        location_dict[address] = dev_path
      if model:
        model_dict[address] = model
  disk_info = _get_disk_info()
  _add_disk_info(address_to_usb_info_dict, disk_info)
  _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict)
  return address_to_usb_info_dict


def get_pyudev_list_of_devices(subsystem=None, devtype=None):
  """Wrapper around pyudev.list_devices.

  Args:
      subsystem (str): maps to pyudev.Context().list_devices arg.
      devtype (str): maps to pyudev.Context().list_devices arg DEVTYPE.

  Returns:
      list: a list of pyudev devices.

  Raises:
      RuntimeError: if platform is not linux.
  """
  args = {}
  if subsystem:
    args['subsystem'] = subsystem
  if devtype:
    args['DEVTYPE'] = devtype
  if sys.platform == 'darwin':
    raise RuntimeError('This method is only supported on linux.')
  context = pyudev.Context()
  return list(context.list_devices(**args))


def _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict):
  """Add cambrionix port number and parent address to each cambrionix child."""
  for cam_address, model in model_dict.items():
    port_mapping = usb_config.CAMBRIONIX_PORT_MAP[model]
    for child_address in address_to_usb_info_dict[cam_address].child_addresses:
      if child_address in address_to_usb_info_dict:
        child_entry = address_to_usb_info_dict[child_address]
        child_entry.usb_hub_address = cam_address
        child_entry.usb_hub_port = _get_cambrionix_port_number(
            location_dict[child_address], port_mapping)


def _add_disk_info(address_to_usb_info_dict, disk_info):
  """Add disk path to the entries that have them."""
  for entry in address_to_usb_info_dict.values():
    serial = entry.serial_number
    if serial in disk_info:
      entry.disk = disk_info[serial]


def _get_address(udev_device):
  """Gets the serial port symlink under '/dev/serial/by-id', mounted drives, or adb serial."""
  devlinks_str = udev_device.get('DEVLINKS', '')
  paths = devlinks_str.split()
  if not devlinks_str:  # Get mounted drives without corresponding symlinks
    if udev_device.get('ID_VENDOR') not in REPEAT_VENDORS:
      return udev_device.get('DEVNAME')

  for path in paths:
    if PERSISTENT_SERIAL_PATH_FOLDER in path:
      return path
    if MOUNTED_DISK_FOLDER in path:
      match = re.match(ADB_SERIAL_REGEX, path)
      if match:
        return match.group(1)
      return path

  return ''


def _get_cambrionix_model(udev_device):
  """Gets the model of the cambrionix."""
  parent = udev_device.parent.parent.parent.parent
  if udev_device.get('ID_MODEL') in usb_config.CAMBRIONIX_PORT_MAP:
    return udev_device.get('ID_MODEL')
  else:
    return parent.get('ID_MODEL')


def _get_cambrionix_port_number(dev_path, port_mapping):
  """Get port from dev path.

  Args:
      dev_path (str): udev device dev path.
      port_mapping (dict): dict of port identifiers to port numbers.

  Returns:
      int: port number or None if not found

  Note:
      Port identifier is the last two digits before the ':'
      So for "/devices/pci0000:00/0000:00:1a.0/usb1/"
             "1-1/1-1.5/1-1.5.4/1-1.5.4.3/1-1.5.4.3:1.3/"
             "ttyUSB42/tty/ttyUSB42"
      "1-1.5.4.3:1.3" is the relevant number.
      '1.3' indicates the FTDI interface
      '4.3' indicates the Cambrionix port
      1 - 1.5 indicates the parent usb.
  """
  match = re.search(IDENTIFIER_REGEX, dev_path)
  if match:
    port_identifier = match.group(1)
    if port_identifier in port_mapping:
      return port_mapping[port_identifier]
  return None


def _get_child_addresses(udev_device):
  """Get the addresses for the relevant children of the udev device."""
  all_children = _get_cambrionix_udev_children(udev_device)
  relevant_children = [
      device for device in all_children
      if device.get('DEVLINKS') and _get_address(device)
  ]
  if udev_device in relevant_children:
    relevant_children.remove(udev_device)
  grand_children = []
  for child_device in relevant_children:  # Remove children of other cambrionixes
    if _get_product_name(child_device) in usb_config.CAMBRIONIX_NAMES:
      grand_children += _get_cambrionix_udev_children(child_device)
      grand_children.remove(child_device)
  return [
      _get_address(child)
      for child in relevant_children
      if child not in grand_children
  ]


def _get_disk_info():
  """Returns a dictionary of disk paths to serial numbers."""
  devices = get_pyudev_list_of_devices(subsystem='block')
  mounted_info = {
      device.get('ID_SERIAL_SHORT', ''): device.get('DEVNAME')
      for device in devices
  }
  return mounted_info


def _get_product_name(udev_device):
  """Gets the product name in a system agnostic way."""
  model = udev_device.get('ID_MODEL', '').replace('_', ' ')
  if model == ANDROID_DEVICE_PRODUCT_NAME:  # Actually an Android device.
    model = 'Android'
  return model


def _process_udev_device(udev_device):
  """Convert pyudev dict to the system agnostic form."""
  model = None
  dev_path = udev_device.get('DEVPATH', '')

  entry = usb_config.UsbInfo()
  address = _get_address(udev_device)
  if not address:
    return '', {}, None, None
  entry.address = address
  entry.product_id = udev_device.get('ID_MODEL_ID', '')
  entry.vendor_id = udev_device.get('ID_VENDOR_ID', '')
  entry.ftdi_interface = int(udev_device.get('ID_USB_INTERFACE_NUM', 0))
  entry.manufacturer = udev_device.get('ID_VENDOR', '')
  entry.serial_number = udev_device.get('ID_SERIAL_SHORT', '')
  entry.product_name = _get_product_name(udev_device)
  if entry.product_name in usb_config.CAMBRIONIX_NAMES:
    entry.child_addresses = _get_child_addresses(udev_device)
    model = _get_cambrionix_model(udev_device)
  return address, entry, dev_path, model


def _get_cambrionix_udev_children(udev_device):
  """Gets all the udev devices representing a cambrionix's children."""
  return list(udev_device.parent.parent.parent.parent.parent.children)
